local imgd = love.image.newImageData('room.png')
local imgw, imgh = imgd:getDimensions()

-- Get a copy, to use a different algorithm on it and compare the results
local img2 = imgd:clone()

local image_orig = love.graphics.newImage(imgd)


local function fillable(x, y)
  if x < 0 or x >= imgw or y < 0 or y >= imgh then
    return false
  end
  -- it's fillable if alpha is zero
  return select(4, imgd:getPixel(x, y)) == 0
end


-- Recursive fill algorithm
local function recursive_fill(x, y)
  if not fillable(x, y) then return end
  imgd:setPixel(x, y, 0, 1, 1, 1)
  recursive_fill(x-1, y)
  recursive_fill(x, y-1)
  recursive_fill(x, y+1)
  recursive_fill(x+1, y)
end

-- Scanline-based. We use a stack instead of a queue because it's easier in Lua.
local function scanline_fill(xorig, yorig)
  local t = {xorig, yorig}
  local tlen = 2
  repeat
    if tlen == 0 then
      return
    end
    local x, y = t[tlen - 1], t[tlen]
    tlen = tlen - 2
    if fillable(x, y) then
      -- Find first fillable pixel to the left
      repeat
        x = x - 1
      until not fillable(x, y)
      x = x + 1

      local prevFillableUp = false
      local prevFillableDn = false

      -- Check if there's a transition from non-fillable to fillable
      -- in the rows above and below. The "previous" pixel is initialized
      -- as non-fillable so that if the first pixel found is fillable,
      -- it counts as a transition and will be enqueued. This is a hack
      -- to simplify the algorithm for the first pixel in the row.
      repeat
        imgd:setPixel(x, y, 0, 1, 1, 1)
        local newFillable = fillable(x, y - 1)
        if not prevFillableUp and newFillable then
          -- Transition from not-fillable to fillable - add to queue
          t[tlen + 1] = x
          t[tlen + 2] = y - 1
          tlen = tlen + 2
        end
        prevFillableUp = newFillable

        -- Repeat for the pixel below
        newFillable = fillable(x, y + 1)
        if not prevFillableDn and newFillable then
          t[tlen + 1] = x
          t[tlen + 2] = y + 1
          tlen = tlen + 2
        end
        prevFillableDn = newFillable

        x = x + 1 -- Next pixel in scanline
      until not fillable(x, y)
    end
  until false
end


recursive_fill(22, 12)

image_recursive = love.graphics.newImage(imgd)

imgd = img2

scanline_fill(22, 12)

image_scanline = love.graphics.newImage(imgd)

love.graphics.setBackgroundColor(1, 1, 1)

function love.draw()
  love.graphics.setColor(0, 0, 0)
  love.graphics.print("Original", 100, 20)
  love.graphics.print("Recursive", 100, 120)
  love.graphics.print("Scanline", 100, 220)
  love.graphics.setColor(1, 1, 1)
  love.graphics.draw(image_orig, 100, 40)
  love.graphics.draw(image_recursive, 100, 140)
  love.graphics.draw(image_scanline, 100, 240)
end

function love.keypressed(k) return k == "escape" and love.event.quit() end
